from typing import Optional, TextIO, Union
import ast

from alang.langs.language import Language, register_language
from alang.langs.writer import CodeWriter
import alang.nodes as nodes
import alang.exprs as exprs
import alang.typs as typs
import alang.funcs as funcs
import alang.stmts as stmts

Code = str

class AWriter(CodeWriter):
    def __init__(self, out: Union[str, TextIO], options: Optional["CodeOptions"]): # type: ignore
        super().__init__(out, options)

    def write_alias(self, a: "Alias"): # type: ignore
        self.write(f"{a.name} = ")
        self.write_type_ref(a.aliased_type)
        self.write("\n")

    def write_binop(self, b: exprs.Binop):
        self.write("(")
        self.write_expr(b.left)
        self.write(" ")
        self.write(b.operator.op)
        self.write(" ")
        self.write_expr(b.right)
        self.write(")")

    def write_constant(self, c: "Constant"): # type: ignore
        self.write(repr(c.value))

    def write_index(self, i: "Index"): # type: ignore
        self.write_expr(i.base)
        self.write("[")
        for ri, r in enumerate(i.ranges):
            if ri > 0:
                self.write(", ")
            if r is None:
                self.write(":")
            else:
                self.write_expr(r)
        self.write("]")

    def write_line_comment(self, comment: str):
        self.write("# ")
        self.write(comment)
        self.write("\n")

    def write_loop(self, f: stmts.Loop):
        self.write(f"for (var {f.var}: ")
        self.write_type_ref(typs.int_type)
        self.write(f" = 0; {f.var} < ")
        self.write_expr(f.count)
        self.write(f"; {f.var}++) {{\n")
        for s in f.statements:
            self.write_node(s)
        self.write("}\n")

    def write_struct(self, s: typs.Struct):
        self.write(f"struct {s.name}:\n")
        for field in s.fields:
            self.write(f"    {field.name}: ")
            self.write_type_ref(field.field_type)
            self.write("\n")

    def write_function(self, f: funcs.Function):
        self.write(f"{f.name}(")
        ps = f.parameters
        for i, param in enumerate(ps):
            self.write(param.name)
            self.write(": ")
            self.write_type_ref(param.parameter_type)
            if i < len(ps) - 1:
                self.write(", ")
        self.write("): ")
        self.write_type_ref(f.return_type)
        self.write(" =")
        ss = f.statements
        if len(ss) == 0:
            self.write(" ...\n")
        elif len(ss) == 1 and ss[0].node_type == nodes.NodeType.RETURN:
            ret_val = ss[0].value
            if ret_val is None:
                self.write(" ()\n")
            else:
                self.write(" ")
                self.write_expr(ret_val)
                self.write("\n")
        else:
            for s in f.statements:
                self.write_node(s)

    def write_return(self, r: stmts.Return):
        self.write("return")
        if r.value is not None:
            self.write(" ")
            self.write_expr(r.value)

    def write_set(self, s: stmts.Set):
        self.write_expr(s.target)
        self.write(" = ")
        self.write_expr(s.value)
        self.write("\n")

    def write_support_node(self, n: "Node"): # type: ignore
        # Support not needed for A
        pass

    def write_name(self, n: exprs.Name):
        self.write(n.name)

    def write_type_ref(self, t: typs.Type):
        self.write(t.name)

class ALanguage(Language):
    def __init__(self):
        super().__init__("a")

    def open_writer(self, out: Union[str, TextIO], options: Optional["CodeOptions"]) -> AWriter: # type: ignore
        return AWriter(out, options)

    def parse_expr(self, expr: Code):
        if expr is None:
            return None
        if isinstance(expr, nodes.Node):
            return expr
        if isinstance(expr, str):
            ast_expr = ast.parse(expr).body[0].value
            return python_expr_to_alang_expr(ast_expr)
        return exprs.Constant(expr)

def python_expr_to_alang_expr(expr: ast.expr):
    if isinstance(expr, ast.Name):
        return exprs.Name(expr.id)
    elif isinstance(expr, ast.BinOp):
        op_name = expr.op.__class__.__name__.lower()
        left = python_expr_to_alang_expr(expr.left)
        right = python_expr_to_alang_expr(expr.right)
        if op_name == "add":
            return left + right
        elif op_name == "sub":
            return left - right
        elif op_name == "mult":
            return left * right
        elif op_name == "div":
            return left / right
        elif op_name == "matmult":
            op_name = "matmul"
        return exprs.Binop(left, op_name, right)
    elif isinstance(expr, ast.Compare):
        if len(expr.ops) != 1:
            raise NotImplementedError(f"Unsupported number of comparison operators: {len(expr.ops)}")
        if len(expr.comparators) != 1:
            raise NotImplementedError(f"Unsupported number of comparators: {len(expr.comparators)}")
        op = expr.ops[0]
        if isinstance(op, ast.Eq):
            op_name = "=="
        elif isinstance(op, ast.NotEq):
            op_name = "!="
        elif isinstance(op, ast.Lt):
            op_name = "<"
        elif isinstance(op, ast.LtE):
            op_name = "<="
        elif isinstance(op, ast.Gt):
            op_name = ">"
        elif isinstance(op, ast.GtE):
            op_name = ">="
        else:
            raise NotImplementedError(f"Unsupported comparison operator: {type(op)}")
        left = python_expr_to_alang_expr(expr.left)
        right = python_expr_to_alang_expr(expr.comparators[0])
        return exprs.Binop(left, op_name, right)
            
    elif isinstance(expr, ast.Constant):
        return exprs.Constant(expr.value)
    else:
        raise NotImplementedError(f"Unsupported expression type: {type(expr)}")

a_lang = ALanguage()
register_language(a_lang)



from typing import Optional

from alang.nodes import Block, Node, NodeAttr, NodeLink, NodeLinks, NodeType
import alang.typs as typs

Code = str

class Function(Block):
    name = NodeAttr()
    parameters = NodeLinks()
    return_type = NodeLink()
    stage = NodeAttr()
    workgroup_size = NodeAttr()

    def __init__(self, name: str, return_type: Optional[typs.Type], *parameters: "Parameter"):
        super().__init__(NodeType.FUNCTION, can_define_types=False, can_define_functions=False, can_define_variables=True, can_define_statements=True)
        self.name = name
        self.return_type = typs.try_resolve_type(return_type, None)
        for p in parameters:
            if isinstance(p, Parameter):
                self.link(p, "parameters")
            else:
                self.param(p)

    def get_default_address_space(self) -> Optional[str]:
        return "function"

    def param(self, name: str, param_type: str = None, location: Optional[str] = None) -> "Function":
        if type(name) == tuple:
            name, param_type = name
        p = Parameter(name, parameter_type=typs.try_resolve_type(param_type, self), location=location)
        self.link(p, "parameters")
        return self
    
    def resolve_type(self, diags: "compiler.Diagnostics") -> typs.Type: # type: ignore
        if self.return_type is None:
            return None
        return_type = self.return_type.resolved_type
        if return_type is None:
            return None
        pts = []
        for p in self.parameters:
            pt = p.resolved_type
            if pt == None:
                return None
            pts.append(pt)
        return typs.FunctionType(return_type, pts)

class Parameter(Node):
    name = NodeAttr()
    parameter_type = NodeLink()
    location = NodeAttr()

    def __init__(self, name: str, parameter_type: "Type" = None, location: Optional[int] = None): # type: ignore
        super().__init__(NodeType.PARAMETER)
        self.name = name
        self.parameter_type = typs.try_resolve_type(parameter_type, None)
        self.location = location

    def resolve_type(self, diags: "compiler.Diagnostics") -> typs.Type: # type: ignore
        pt = self.parameter_type
        if pt is not None:
            return pt.resolved_type



from alang.exprs import Constant

def test_constant_mul():
    assert Constant(42).mul(Constant(69)).value == 42 * 69
    assert (Constant(42) * Constant(69)).value == 42 * 69
    assert (Constant(42) * 69).value == 42 * 69
    assert (Constant(42) * "23 * 3").value == 42 * 69

def test_constant_add():
    assert Constant(42).add(Constant(69)).value == 42 + 69
    assert (Constant(42) + Constant(69)).value == 42 + 69
    assert (Constant(42) + 69).value == 42 + 69
    assert (Constant(42) + "60 + 9").value == 42 + 69

def test_constant_sub():
    assert Constant(42).sub(Constant(69)).value == 42 - 69
    assert (Constant(42) - Constant(69)).value == 42 - 69
    assert (Constant(42) - 69).value == 42 - 69
    assert (Constant(42) - "72 - 3").value == 42 - 69
